﻿using System.IO;
using System.Windows;
using System.Windows.Media;
using Microsoft.Kinect;
using System.Windows.Shapes;
using System.Windows.Controls;
using System;
using Microsoft.Samples.Kinect.SwipeGestureRecognizer;



namespace KinectPaint
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        // Set to true to activate the smooth skeleton filtering
        private static bool ACTIVATE_SMOOTHING_FILTER = true;

        // Set to true for showing the currently tracked Kinect data
        private static bool SHOW_KINECT_SKELETON = true;

        // Set to true for showing the kinect tracking info label
        private static bool SHOW_KINECT_TRACKING_INFO = true;

        // Set to true for using the swipe gesture recognizer
        private static bool USE_GESTURE_RECOGNIZER = true;


        // The size of our drawing brushes
        private double drawingSize = 50;

        // Current drawing color
        private Color drawingColor = Colors.Red;



        /// <summary>
        /// Incoming tracking data from the Kinect
        /// </summary>
        /// <param name="skeletons"></param>
        private void ProcessTrackingData(Skeleton[] skeletons)
        {
            // Iterate over all currently tracked skeletons. The array is empty in case
            // no person is currently visible to the Kinect camera
            foreach (Skeleton skeleton in skeletons)
            {
                if (skeleton.TrackingState == SkeletonTrackingState.Tracked)
                {
                    // Convert 3D points to pixels on the screen
                    Point shoulderPosition = this.SkeletonPointToScreen(skeleton.Joints[JointType.ShoulderCenter].Position);
                    Point leftHandPosition = this.SkeletonPointToScreen(skeleton.Joints[JointType.HandLeft].Position);
                    Point rightHandPosition = this.SkeletonPointToScreen(skeleton.Joints[JointType.HandRight].Position);

                    // Calculate distance of person to the screen
                    double distanceToScreen = skeleton.Joints[JointType.ShoulderCenter].Position.Z * 100;
                    this.labelInfo.Content = "Distance to screen: " + distanceToScreen.ToString("0") + " cm";



                    // Calculate Z-distance of left hand relative to body
                    double distanceLeftHandToBody = 
                        (skeleton.Joints[JointType.ShoulderCenter].Position.Z -
                        skeleton.Joints[JointType.HandLeft].Position.Z) * 100; 

                    // Draw only of hand is streched out at least 40 cm from the body
                    if (distanceLeftHandToBody > 40) this.DrawCircle(leftHandPosition, new SolidColorBrush(this.drawingColor), 1.0);


                    // Calculate Z-distance of right hand relative to body
                    double distanceRightHandToBody =
                        (skeleton.Joints[JointType.ShoulderCenter].Position.Z -
                        skeleton.Joints[JointType.HandRight].Position.Z) * 100; 

                    // Draw only of hand is streched out at least 40 cm from the body
                    if (distanceRightHandToBody > 40) this.DrawCircle(rightHandPosition, new SolidColorBrush(this.drawingColor), 1.0);

                }
            }
        }



        /// <summary>
        /// Event callback for swipe gesture with left hand
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void recognizer_SwipeLeftDetected(object sender, KinectGestureEventArgs e)
        {
            Dispatcher.BeginInvoke(new Action(() =>
            {
                
            }));
        }



        /// <summary>
        /// Event callback for swipe gesture with right hand
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void recognizer_SwipeRightDetected(object sender, KinectGestureEventArgs e)
        {
            Dispatcher.BeginInvoke(new Action(() =>
            {
                
            }));
        }









        #region Utility functions

        /// <summary>
        /// Calculates orientation of a person's body to the display
        /// </summary>
        private double CalculateBodyOrientation(Skeleton skeleton)
        {
            SkeletonPoint shoulderLeft3D = skeleton.Joints[JointType.ShoulderLeft].Position;
            SkeletonPoint shoulderRight3D = skeleton.Joints[JointType.ShoulderRight].Position;

            Point shoulderLeft = new Point(shoulderLeft3D.X, shoulderLeft3D.Z);
            Point shoulderRight = new Point(shoulderRight3D.X, shoulderRight3D.Z);

            Line2D line1 = new Line2D(shoulderLeft, shoulderRight);
            Line2D line2 = new Line2D(shoulderLeft, new Point(shoulderLeft.X + 1, shoulderLeft.Y));

            double orientation = RadianToDegree(AngleBetween2Lines(line1, line2));
            if (orientation < -180.0) orientation = 360.0 + orientation;
            return orientation;
        }

        /// <summary>
        /// Calculate the angle between two lines in 2D
        /// </summary>
        public static double AngleBetween2Lines(Line2D line1, Line2D line2)
        {
            double angle1 = Math.Atan2(line1.P1.Y - line1.P2.Y, line1.P1.X - line1.P2.X);
            double angle2 = Math.Atan2(line2.P1.Y - line2.P2.Y, line2.P1.X - line2.P2.X);
            return angle1 - angle2;
        }

        /// <summary>
        /// Representing a line in 2D
        /// </summary>
        public class Line2D
        {
            public Point P1 = new Point();
            public Point P2 = new Point();
            public Line2D(Point p1, Point p2) { this.P1 = p1; this.P2 = p2; }
        }

        /// <summary>
        /// Convert angles in degrees to radian
        /// </summary>
        private static double DegreeToRadian(double angle) { return Math.PI * angle / 180.0; }

        /// <summary>
        /// Convert angles in radian to degrees
        /// </summary>
        private static double RadianToDegree(double angle) { return angle * (180.0 / Math.PI); }


        /// <summary>
        /// Distance between two points in 3D
        /// </summary>
        public static double Distance3D(double x1, double y1, double z1, double x2, double y2, double z2)
        {
            return (double)Math.Sqrt(Math.Pow((x2 - x1), 2) + Math.Pow((y2 - y1), 2) + Math.Pow((z2 - z1), 2));
        }

        /// <summary>
        /// Distance between two skeleton points
        /// </summary>
        /// <param name="sp1"></param>
        /// <param name="sp2"></param>
        /// <returns></returns>
        public static double DistanceBetweenSkeletonPoints(SkeletonPoint sp1, SkeletonPoint sp2)
        {
            return Distance3D(sp1.X, sp1.Y, sp1.Z, sp2.X, sp2.Y, sp2.Z);
        }

        /// <summary>
        ///  Draw a circle on screen
        /// </summary>
        private void DrawCircle(Point position, Brush color, double opacity)
        {
            Ellipse ellipse = new Ellipse();
            ellipse.Fill = color;
            ellipse.Width = drawingSize;
            ellipse.Height = drawingSize;
            ellipse.Opacity = opacity;
            ellipse.SetValue(Canvas.LeftProperty, (double)(position.X - drawingSize / 2.0));
            ellipse.SetValue(Canvas.TopProperty, (double)(position.Y - drawingSize / 2.0));
            this.canvas.Children.Add(ellipse);
        }

        /// <summary>
        /// Random number generator
        /// </summary>
        private Random random = new Random();

        /// <summary>
        ///  Get a random color
        /// </summary>
        private Color GetRandomColour()
        {
            return Color.FromArgb(
                (byte)255,
                (byte)this.random.Next(255),
                (byte)this.random.Next(255),
                (byte)this.random.Next(255));
        }


        /// <summary>
        /// The recognizer being used.
        /// </summary>
        private readonly Recognizer activeRecognizer;

        /// <summary>
        /// Create gesture recognizer for swipe gestures
        /// </summary>
        /// <returns>The wired-up recognizer.</returns>
        private Recognizer CreateRecognizer()
        {
            // Instantiate a recognizer
            var recognizer = new Recognizer();

            // Register for swipe gesture events from left ot RIGHT
            recognizer.SwipeRightDetected += new EventHandler<KinectGestureEventArgs>(recognizer_SwipeRightDetected);

            // Register for swipe gestures from right to LEFT
            recognizer.SwipeLeftDetected += new EventHandler<KinectGestureEventArgs>(recognizer_SwipeLeftDetected);

            return recognizer;
        }

        #endregion



        #region Members for Kinect skeleton processing and drawing

        /// <summary>
        /// Width of output drawing
        /// </summary>
        private const float RenderWidth = 640.0f;

        /// <summary>
        /// Height of our output drawing
        /// </summary>
        private const float RenderHeight = 480.0f;

        /// <summary>
        /// Thickness of drawn joint lines
        /// </summary>
        private const double JointThickness = 3;

        /// <summary>
        /// Thickness of body center ellipse
        /// </summary>
        private const double BodyCenterThickness = 10;

        /// <summary>
        /// Thickness of clip edge rectangles
        /// </summary>
        private const double ClipBoundsThickness = 10;

        /// <summary>
        /// Brush used to draw skeleton center point
        /// </summary>
        private readonly Brush centerPointBrush = Brushes.Blue;

        /// <summary>
        /// Brush used for drawing joints that are currently tracked
        /// </summary>
        //private readonly Brush trackedJointBrush = new SolidColorBrush(Color.FromArgb(255, 68, 192, 68));
        private readonly Brush trackedJointBrush = Brushes.DarkGray;

        /// <summary>
        /// Brush used for drawing joints that are currently inferred
        /// </summary>        
        //private readonly Brush inferredJointBrush = Brushes.Yellow;
        private readonly Brush inferredJointBrush = Brushes.DarkGray;

        /// <summary>
        /// Pen used for drawing bones that are currently tracked
        /// </summary>
        //private readonly Pen trackedBonePen = new Pen(Brushes.Green, 6);
        private readonly Pen trackedBonePen = new Pen(Brushes.DarkGray, 6);

        /// <summary>
        /// Pen used for drawing bones that are currently inferred
        /// </summary>        
        private readonly Pen inferredBonePen = new Pen(Brushes.DarkGray, 1);

        /// <summary>
        /// Active Kinect sensor
        /// </summary>
        private KinectSensor sensor;

        /// <summary>
        /// Drawing group for skeleton rendering output
        /// </summary>
        private DrawingGroup drawingGroup;

        /// <summary>
        /// Drawing image that we will display
        /// </summary>
        private DrawingImage imageSource;

        #endregion



        #region Methods for Kinect processing

        /// <summary>
        /// Initializes a new instance of the MainWindow class.
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            if (!SHOW_KINECT_SKELETON)
            {
                this.viewboxKinect.Visibility = System.Windows.Visibility.Collapsed;
            }

            if (!SHOW_KINECT_TRACKING_INFO)
            {
                this.canvasInfo.Visibility = System.Windows.Visibility.Collapsed;
            }

            // Create gesture recognizer
            this.activeRecognizer = this.CreateRecognizer();
        }


        /// <summary>
        /// Execute startup tasks
        /// </summary>
        /// <param name="sender">object sending the event</param>
        /// <param name="e">event arguments</param>
        private void WindowLoaded(object sender, RoutedEventArgs e)
        {
            // Create the drawing group we'll use for drawing
            this.drawingGroup = new DrawingGroup();

            // Create an image source that we can use in our image control
            this.imageSource = new DrawingImage(this.drawingGroup);

            // Display the drawing using our image control
            Image.Source = this.imageSource;

            // Look through all sensors and start the first connected one.
            // This requires that a Kinect is connected at the time of app startup.
            // To make your app robust against plug/unplug, 
            // it is recommended to use KinectSensorChooser provided in Microsoft.Kinect.Toolkit
            foreach (var potentialSensor in KinectSensor.KinectSensors)
            {
                if (potentialSensor.Status == KinectStatus.Connected)
                {
                    this.sensor = potentialSensor;
                    break;
                }
            }

            if (null != this.sensor)
            {
                TransformSmoothParameters parameters = new TransformSmoothParameters
                {
                    Smoothing = 0.7f,
                    Correction = 0.3f,
                    Prediction = 0.4f,
                    JitterRadius = 1.0f,
                    MaxDeviationRadius = 0.5f
                };

                if (ACTIVATE_SMOOTHING_FILTER)
                {
                    // Turn on the skeleton stream to receive skeleton frames
                    this.sensor.SkeletonStream.Enable(parameters);
                }
                else
                {
                    // Turn on the skeleton stream to receive skeleton frames
                    this.sensor.SkeletonStream.Enable();
                }

                // Add an event handler to be called whenever there is new color frame data
                this.sensor.SkeletonFrameReady += this.SensorSkeletonFrameReady;

                // Start the sensor!
                try
                {
                    this.sensor.Start();
                }
                catch (IOException)
                {
                    this.sensor = null;
                }
            }

        }

        /// <summary>
        /// Execute shutdown tasks
        /// </summary>
        /// <param name="sender">object sending the event</param>
        /// <param name="e">event arguments</param>
        private void WindowClosing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            if (null != this.sensor)
            {
                this.sensor.Stop();
            }
        }

        /// <summary>
        /// Event handler for Kinect sensor's SkeletonFrameReady event
        /// </summary>
        /// <param name="sender">object sending the event</param>
        /// <param name="e">event arguments</param>
        private void SensorSkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            Skeleton[] skeletons = new Skeleton[0];

            using (SkeletonFrame skeletonFrame = e.OpenSkeletonFrame())
            {
                if (skeletonFrame != null)
                {
                    skeletons = new Skeleton[skeletonFrame.SkeletonArrayLength];
                    skeletonFrame.CopySkeletonDataTo(skeletons);

                    if (USE_GESTURE_RECOGNIZER)
                    {
                        this.activeRecognizer.Recognize(sender, skeletonFrame, skeletons);
                    }
                }
            }

            using (DrawingContext dc = this.drawingGroup.Open())
            {
                // Draw a transparent background to set the render size
                dc.DrawRectangle(Brushes.Black, null, new Rect(0.0, 0.0, RenderWidth, RenderHeight));

                if (skeletons.Length != 0)
                {
                    foreach (Skeleton skel in skeletons)
                    {
                        if (skel.TrackingState == SkeletonTrackingState.Tracked)
                        {
                            this.DrawBonesAndJoints(skel, dc);
                        }
                        else if (skel.TrackingState == SkeletonTrackingState.PositionOnly)
                        {
                            dc.DrawEllipse(
                            this.centerPointBrush,
                            null,
                            this.SkeletonPointToScreenFixed(skel.Position),
                            BodyCenterThickness,
                            BodyCenterThickness);
                        }
                    }
                }

                // prevent drawing outside of our render area
                this.drawingGroup.ClipGeometry = new RectangleGeometry(new Rect(0.0, 0.0, RenderWidth, RenderHeight));

                this.ProcessTrackingData(skeletons);
            }
        }

        /// <summary>
        /// Draws a skeleton's bones and joints
        /// </summary>
        /// <param name="skeleton">skeleton to draw</param>
        /// <param name="drawingContext">drawing context to draw to</param>
        private void DrawBonesAndJoints(Skeleton skeleton, DrawingContext drawingContext)
        {
            // Render Torso
            this.DrawBone(skeleton, drawingContext, JointType.Head, JointType.ShoulderCenter);
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderCenter, JointType.ShoulderLeft);
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderCenter, JointType.ShoulderRight);
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderCenter, JointType.Spine);
            this.DrawBone(skeleton, drawingContext, JointType.Spine, JointType.HipCenter);
            this.DrawBone(skeleton, drawingContext, JointType.HipCenter, JointType.HipLeft);
            this.DrawBone(skeleton, drawingContext, JointType.HipCenter, JointType.HipRight);

            // Left Arm
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderLeft, JointType.ElbowLeft);
            this.DrawBone(skeleton, drawingContext, JointType.ElbowLeft, JointType.WristLeft);
            this.DrawBone(skeleton, drawingContext, JointType.WristLeft, JointType.HandLeft);

            // Right Arm
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderRight, JointType.ElbowRight);
            this.DrawBone(skeleton, drawingContext, JointType.ElbowRight, JointType.WristRight);
            this.DrawBone(skeleton, drawingContext, JointType.WristRight, JointType.HandRight);

            // Left Leg
            this.DrawBone(skeleton, drawingContext, JointType.HipLeft, JointType.KneeLeft);
            this.DrawBone(skeleton, drawingContext, JointType.KneeLeft, JointType.AnkleLeft);
            this.DrawBone(skeleton, drawingContext, JointType.AnkleLeft, JointType.FootLeft);

            // Right Leg
            this.DrawBone(skeleton, drawingContext, JointType.HipRight, JointType.KneeRight);
            this.DrawBone(skeleton, drawingContext, JointType.KneeRight, JointType.AnkleRight);
            this.DrawBone(skeleton, drawingContext, JointType.AnkleRight, JointType.FootRight);
 
            // Render Joints
            foreach (Joint joint in skeleton.Joints)
            {
                Brush drawBrush = null;

                if (joint.TrackingState == JointTrackingState.Tracked)
                {
                    drawBrush = this.trackedJointBrush;                    
                }
                else if (joint.TrackingState == JointTrackingState.Inferred)
                {
                    drawBrush = this.inferredJointBrush;                    
                }

                if (drawBrush != null)
                {
                    drawingContext.DrawEllipse(drawBrush, null, this.SkeletonPointToScreenFixed(joint.Position), JointThickness, JointThickness);
                }
            }
        }

        /// <summary>
        /// Maps a SkeletonPoint to lie within our render space and converts to Point
        /// </summary>
        /// <param name="skelpoint">point to map</param>
        /// <returns>mapped point</returns>
        private Point SkeletonPointToScreenFixed(SkeletonPoint skelpoint)
        {
            // Convert point to depth space.  
            // We are not using depth directly, but we do want the points in our 640x480 output resolution.
            DepthImagePoint depthPoint = this.sensor.MapSkeletonPointToDepth(
                                                                             skelpoint,
                                                                             DepthImageFormat.Resolution640x480Fps30);
            return new Point(depthPoint.X, depthPoint.Y);
        }

        /// <summary>
        /// Maps a SkeletonPoint to lie within our render space and converts to Point
        /// </summary>
        /// <param name="skelpoint">point to map</param>
        /// <returns>mapped point</returns>
        private Point SkeletonPointToScreen(SkeletonPoint skelpoint)
        {
            // Convert point to depth space.  
            // We are not using depth directly, but we do want the points in our 640x480 output resolution.
            DepthImagePoint depthPoint = this.sensor.MapSkeletonPointToDepth(
                                                                             skelpoint,
                                                                             DepthImageFormat.Resolution640x480Fps30);
            return new Point(depthPoint.X * (this.canvas.ActualWidth / 640.0), depthPoint.Y * (this.canvas.ActualHeight / 480.0));
        }

        /// <summary>
        /// Draws a bone line between two joints
        /// </summary>
        /// <param name="skeleton">skeleton to draw bones from</param>
        /// <param name="drawingContext">drawing context to draw to</param>
        /// <param name="jointType0">joint to start drawing from</param>
        /// <param name="jointType1">joint to end drawing at</param>
        private void DrawBone(Skeleton skeleton, DrawingContext drawingContext, JointType jointType0, JointType jointType1)
        {
            Joint joint0 = skeleton.Joints[jointType0];
            Joint joint1 = skeleton.Joints[jointType1];

            // If we can't find either of these joints, exit
            if (joint0.TrackingState == JointTrackingState.NotTracked ||
                joint1.TrackingState == JointTrackingState.NotTracked)
            {
                return;
            }

            // Don't draw if both points are inferred
            if (joint0.TrackingState == JointTrackingState.Inferred &&
                joint1.TrackingState == JointTrackingState.Inferred)
            {
                return;
            }

            // We assume all drawn bones are inferred unless BOTH joints are tracked
            Pen drawPen = this.inferredBonePen;
            if (joint0.TrackingState == JointTrackingState.Tracked && joint1.TrackingState == JointTrackingState.Tracked)
            {
                drawPen = this.trackedBonePen;
            }

            drawingContext.DrawLine(drawPen, this.SkeletonPointToScreenFixed(joint0.Position), this.SkeletonPointToScreenFixed(joint1.Position));
        }

      

        #endregion



        #region Key pressed event handler
        private void Window_KeyDown(object sender, System.Windows.Input.KeyEventArgs e)
        {
            if (e.Key == System.Windows.Input.Key.Escape) Application.Current.Shutdown(0);
        }
        #endregion


    }
}